import math
import numpy as np

from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Parameter

from .base_problem import Problem


class QUBO(Problem):
    def __init__(self, Q=None, c=None, b=None) -> None:
        """
        Implements the mapping from the parameters in params to the QUBO problem.
        Is expected to be called by the child class.

        # The QUBO will be on this form:
        # min x^T Q x + c^T x + b
        """
        assert type(Q) is np.ndarray, "Q needs to be a numpy ndarray, but is " + str(
            type(Q)
        )
        assert (
            Q.ndim == 2
        ), "Q needs to be a 2-dimensional numpy ndarray, but has dim " + str(Q.ndim)
        assert Q.shape[0] == Q.shape[1], "Q needs to be a square matrix, but is " + str(
            Q.shape
        )
        n = Q.shape[0]

        # Check if Q is lower triangular
        self.lower_triangular_Q = np.allclose(Q, np.tril(Q))

        self.QUBO_Q = Q

        if c is None:
            c = np.zeros(n)
        assert type(c) is np.ndarray, "c needs to be a numpy ndarray, but is " + str(
            type(c)
        )
        assert (
            c.ndim == 1
        ), "c needs to be a 1-dimensional numpy ndarray, but has dim " + str(Q.ndim)
        assert c.shape[0] == n, (
            "c is of size "
            + str(c.shape[0])
            + " but should be compatible size to Q, meaning "
            + str(n)
        )
        self.QUBO_c = c

        if b is None:
            b = 0.0
        assert np.isscalar(b), "b is expected to be scalar, but is " + str(b)
        self.QUBO_b = b

    def cost(self, string):
        x = np.array(list(map(int, string)))
        return -(x.T @ self.QUBO_Q @ x + self.QUBO_c.T @ x + self.QUBO_b)

    def create_circuit(self):
        if not self.lower_triangular_Q:
            LOG.error("Function not implemented!", func=self.create_circuit.__name__)
            raise NotImplementedError
        self.createParameterizedCostCircuitTril()

    def createParameterizedCostCircuitTril(self):
        """
        Creates a parameterized circuit of the triangularized QUBO problem.
        """
        q = QuantumRegister(self.N_qubits)
        self.circuit = QuantumCircuit(q)
        cost_param = Parameter("x_gamma")

        ### cost Hamiltonian
        for i in range(self.N_qubits):
            w_i = 0.5 * (self.QUBO_c[i] + np.sum(self.QUBO_Q[:, i]))

            if not math.isclose(w_i, 0, abs_tol=1e-7):
                self.circuit.rz(cost_param * w_i, q[i])

            for j in range(i + 1, self.N_qubits):
                w_ij = 0.25 * self.QUBO_Q[j][i]

                if not math.isclose(w_ij, 0, abs_tol=1e-7):
                    self.circuit.cx(q[i], q[j])
                    self.circuit.rz(cost_param * w_ij, q[j])
                    self.circuit.cx(q[i], q[j])

    # def __str2np(self, s):
    #    x = np.array(list(map(int, s)))
    #    assert len(x) == self.N_qubits, (
    #        "bitstring  "
    #        + s
    #        + " of wrong size. Expected "
    #        + str(self.N_qubits)
    #        + " but got "
    #        + str(len(x))
    #    )
    #    return x



import numpy as np
import matplotlib.pyplot as plt
import datetime

from qiskit_finance.data_providers import RandomDataProvider


class AssetData:
    def __init__(
        self, N_assets, num_days=101, seed=0, start_time=datetime.datetime(2020, 1, 1)
    ):
        """
        init function that initializes member variables

        :param params: additional parameters
        """
        self.N = N_assets
        self.num_days = num_days
        self.start_time = start_time
        self.end_time = start_time + datetime.timedelta(self.num_days)

        self.tickers = [("TICKER%s" % i) for i in range(self.N)]
        self.fin_data = RandomDataProvider(
            tickers=self.tickers, start=self.start_time, end=self.end_time, seed=seed
        )
        self.fin_data.run()

        self.cov_matrix = self.fin_data.get_period_return_covariance_matrix()
        self.exp_return = self.fin_data.get_period_return_mean_vector()

    def plotAssets(self, figsize=(12, 4)):
        fig = plt.figure(figsize=figsize)
        gs = fig.add_gridspec(1, 3)
        axs = [None] * 2
        axs[0] = fig.add_subplot(gs[0, 0:2])
        t = [self.start_time + datetime.timedelta(dt) for dt in range(self.num_days)]
        for i, ticker in enumerate(self.tickers):
            axs[0].plot(t, self.fin_data._data[i], label=ticker)
        axs[0].set_xticklabels(axs[0].get_xticklabels(), rotation=-30)
        axs[0].legend()
        axs[0].set_title("time development")

        axs[1] = fig.add_subplot(gs[0, 2])
        im = axs[1].imshow(self.cov_matrix)
        fig.colorbar(im, ax=axs[1], shrink=0.8)
        axs[1].set_title("Period return cov. matrix")

    def plotPeriodReturns(self, figsize=(8, 4)):
        fig = plt.figure(figsize=figsize)
        gs = fig.add_gridspec(1, 3)
        axs = [None] * 2
        axs[0] = fig.add_subplot(gs[0, 0:2])
        t = [self.start_time + datetime.timedelta(dt) for dt in range(self.num_days)]
        for i, ticker in enumerate(self.tickers):
            axs[0].plot(t, self.fin_data._data[i], label=ticker)
        axs[0].set_xticklabels(axs[0].get_xticklabels(), rotation=-30)
        axs[0].legend()
        axs[0].set_title("time development")



from qiskit.circuit import Parameter
from qiskit.circuit.library import PhaseGate

from qaoa.mixers.base_mixer import Mixer
from qaoa.initialstates.base_initialstate import InitialState


class Grover(Mixer):
    def __init__(self, subcircuit: InitialState) -> None:
        """
        Args:
            subcircuit (InitialState): the circuit that is to be tensorised
        """
        self.subcircuit = subcircuit
        self.mixer_param = Parameter("x_beta")

    def create_circuit(self):
        # given feasibel states f \in F,
        # Let US the circuit that prepares US = 1/|F| \sum_{f\inF} |f>
        # The Grover mixer has the form US^\dagger X^n C^{n-1}Phase X^n US,

        self.subcircuit.create_circuit()
        US = self.subcircuit.circuit

        # US^\dagger
        self.circuit = US.inverse()
        # X^n
        self.circuit.x(range(self.subcircuit.N_qubits))
        # C^{n-1}Phase
        phase_gate = PhaseGate(-self.mixer_param).control(self.subcircuit.N_qubits - 1)
        self.circuit.append(phase_gate, self.circuit.qubits)
        # X^n
        self.circuit.x(range(self.subcircuit.N_qubits))
        # US
        self.circuit.compose(US, range(self.subcircuit.N_qubits), inplace=True)



import numpy as np
import os, sys

def load_FR_CR(filename):

    if not os.path.isfile(filename):
        print("File not found.")
        raise IOError

    # Matrix in file on format
    # C_1: index of ones in row 1 separated by comma
    # C_2: index of ones in row 2 ...
    # ...
    # C_R: index of ones in row R
    # best state 

    file = np.loadtxt(filename, dtype = str)

    R,F = filename.split('/')[-1].split('_')[1:3]
    R,F = int(R), int(F)

    FR   = np.zeros((F,R))
    CR   = np.zeros(R)
    best = np.zeros(R)
    
    for r in range(R):
        CR[r]   = float(file[r].split(':')[0])    
        indexes = file[r].split(':')[1].split(',')
        for ind in indexes:
            FR[int(ind), r ] = 1
    
    best_str = file[R].split(',')
    best     = np.array([int(i) for i in best_str]) 
            
    return FR, CR, best

def npy_loader(filename):
    """
    Loading the examples saved on npy format.

    Parameters
    ----------
    filename : string
        filename
    
    Returns
    -------
    FR : array
        Constraint matrix
    CR : array
        Weights

    """
    matrix = np.load(filename)

    # The costs are saved in the last column
    CR     = matrix[-1]
    FR     = matrix[:-1]

    return FR, CR



from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Parameter

from .base_problem import Problem


class MaxCut(Problem):
    def __init__(self, G) -> None:
        self.G = G
        self.N_qubits = self.G.number_of_nodes()

    def cost(self, string):
        C = 0
        for edge in self.G.edges():
            i = int(edge[0])
            j = int(edge[1])
            if string[i] != string[j]:
                w = self.G[edge[0]][edge[1]]["weight"]
                C += w
        return C

    def create_circuit(self):
        """
        Adds a parameterized circuit for the cost part to the member variable self.parameteried_circuit
        and a parameter to the parameter list self.gamma_params
        """
        q = QuantumRegister(self.N_qubits)
        self.circuit = QuantumCircuit(q)
        cost_param = Parameter("x_gamma")

        ### cost Hamiltonian
        for edge in self.G.edges():
            i = int(edge[0])
            j = int(edge[1])
            w = self.G[edge[0]][edge[1]]["weight"]
            wg = w * cost_param
            self.circuit.cx(q[i], q[j])
            self.circuit.rz(wg, q[j])
            self.circuit.cx(q[i], q[j])



from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Parameter
from qiskit.circuit.library import XXPlusYYGate

from qaoa.mixers import Mixer

import math
import itertools

import numpy as np


class XY(Mixer):
    def __init__(self, topology=None) -> None:
        self.topology = topology
        self.mixer_param = Parameter("x_beta")

    def create_circuit(self):
        if not self.topology:
            print('No topology specified for the XY-mixer, assuming "ring" topology')
            self.topology = XY.generate_pairs(self.N_qubits)

        q = QuantumRegister(self.N_qubits)
        self.circuit = QuantumCircuit(q)

        for i, e in enumerate(self.topology):
            self.circuit.append(XXPlusYYGate(0.5 * self.mixer_param), e)

    @staticmethod
    def generate_pairs(n, case="ring"):
        # default ring, otherwise "chain"
        if n < 2:
            return []  # Not enough elements to form any pairs

        pairs = [[i, i + 1] for i in range(n - 1)]

        if case == "ring":
            pairs.append([n - 1, 0])

        return pairs



from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Parameter

from .base_mixer import Mixer


class X(Mixer):
    def __init__(self) -> None:
        self.mixer_param = Parameter("x_beta")

    def create_circuit(self):
        q = QuantumRegister(self.N_qubits)

        self.circuit = QuantumCircuit(q)
        self.circuit.rx(-2 * self.mixer_param, range(self.N_qubits))



import structlog

LOG = structlog.get_logger(file=__name__)

import numpy as np

from qiskit import (
    QuantumCircuit,
    QuantumRegister,
    ClassicalRegister,
    AncillaRegister,
    transpile,
)
from qiskit.circuit import Parameter
from qiskit.primitives import Sampler
from qiskit_algorithms.optimizers import COBYLA

from qiskit_aer import Aer

from qaoa.initialstates import InitialState
from qaoa.mixers import Mixer
from qaoa.problems import Problem
from qaoa.util import Statistic, BitFlip, post_processing


class OptResult:
    """
    Class for results of optimization at depth p
    """

    def __init__(self, depth):
        self.depth = depth

        self.angles = []
        self.Exp = []
        self.Var = []
        self.WorstCost = []
        self.BestCost = []
        self.BestSols = []
        self.shots = []

        self.index_Exp_min = -1

    def add_iteration(self, angles, stat, shots):
        self.angles.append(angles)
        self.Exp.append(-stat.get_CVaR())
        self.Var.append(stat.get_Variance())
        self.BestCost.append(-stat.get_max())
        self.WorstCost.append(-stat.get_min())
        self.BestSols.append(stat.get_max_sols())
        self.shots.append(shots)

    def compute_best_index(self):
        self.index_Exp_min = self.Exp.index(min(self.Exp))

    def get_best_Exp(self):
        return self.Exp[self.index_Exp_min]

    def get_best_Var(self):
        return self.Var[self.index_Exp_min]

    def get_best_angles(self):
        return self.angles[self.index_Exp_min]

    def num_fval(self):
        return len(self.Exp)

    def num_shots(self):
        return sum(self.shots)

    def get_best_solution(self):
        best_cost = np.min(self.BestCost)
        iterations_with_best_cost = np.where(self.BestCost == best_cost)[0]

        all_best_sols = []
        for i in iterations_with_best_cost:
            all_best_sols.append(self.BestSols[i])
        # flatten the list:
        all_best_sols = [item for sublist in all_best_sols for item in sublist]
        best_sols = np.unique(all_best_sols)
        return best_sols, best_cost


class QAOA:
    """Main class"""

    def __init__(
        self,
        problem,
        mixer,
        initialstate,
        backend=Aer.get_backend("qasm_simulator"),
        noisemodel=None,
        optimizer=[COBYLA, {}],  # optimizer, options
        precision=None,
        shots=1024,
        cvar=1,
        memorysize=-1,
        interpolate=True,
        flip=False,
        post=False,
        anc=0,
    ) -> None:
        """
        A QAO-Ansatz consist of these parts:

        :problem of Basetype Problem,
        implementing the phase circuit and the cost.

        :mixer of Basetype Mixer,
        specifying the mixer circuit.

        :initialstate of Basetype InitialState,
        specifying the initial state.

        :backend: backend
        :noisemodel: noisemodel
        :optmizer: optimizer as a list containing [optimizer object, options dict]
        :precision: precision to reach for expectation value based on error=variance/sqrt(shots)
        :shots: if precision=None, the number of samples taken
                      if precision!=None, the minimum number of samples taken
        :cvar: used for CVar
        """

        assert issubclass(type(problem), Problem)
        assert issubclass(type(mixer), Mixer)
        assert issubclass(type(initialstate), InitialState)

        self.problem = problem
        self.mixer = mixer
        self.initialstate = initialstate
        self.initialstate.setNumQubits(self.problem.N_qubits)
        self.mixer.setNumQubits(self.problem.N_qubits)

        self.parameterized_circuit = None
        self.parameterized_circuit_depth = 0

        self.backend = backend
        self.optimizer = optimizer
        self.noisemodel = noisemodel
        self.shots = shots
        self.precision = precision
        self.stat = Statistic(cvar=cvar)
        self.cvar = cvar
        self.memorysize = memorysize
        self.memory = self.memorysize > 0
        self.interpolate = interpolate

        self.usebarrier = False
        self.isQNSPSA = False

        self.current_depth = 0  # depth at which local optimization has been done

        self.gamma_params = None
        self.beta_params = None

        self.Exp_sampled_p1 = None
        self.landscape_p1_angles = {}
        self.Var_sampled_p1 = None
        self.MaxCost_sampled_p1 = None
        self.MinCost_sampled_p1 = None

        self.optimization_results = {}
        self.memory_lists = []

        self.flip = flip
        self.flipper = BitFlip(self.problem.N_qubits)
        self.bitflips = []

        self.post = post
        self.Exp_post_processed = None
        self.Var_post_processed = None
        self.samplecount_hists = {}
        self.last_hist = {}

        self.anc = anc

    def exp_landscape(self):
        ### at depth p = 1
        return self.Exp_sampled_p1

    def var_landscape(self):
        ### at depth p = 1
        return self.Var_sampled_p1

    def get_Exp(self, depth=None):
        if not depth:
            ret = []
            for i in range(1, self.current_depth + 1):
                ret.append(self.optimization_results[i].get_best_Exp())
            return ret
        if depth > self.current_depth + 1:
            raise ValueError
        return self.optimization_results[depth].get_best_Exp()

    def get_Var(self, depth):
        if depth > self.current_depth + 1:
            raise ValueError
        return self.optimization_results[depth].get_best_Var()

    def get_beta(self, depth):
        if depth > self.current_depth + 1:
            raise ValueError
        return self.optimization_results[depth].get_best_angles()[1::2]

    def get_gamma(self, depth):
        if depth > self.current_depth + 1:
            raise ValueError
        return self.optimization_results[depth].get_best_angles()[::2]

    def get_angles(self, depth):
        if depth > self.current_depth + 1:
            raise ValueError
        return self.optimization_results[depth].get_best_angles()

    def get_memory_lists(self):
        return self.memory_lists

    def createParameterizedCircuit(self, depth):
        if self.parameterized_circuit_depth == depth:
            LOG.info(
                "Circuit is already of depth " + str(self.parameterized_circuit_depth)
            )
            return

        self.problem.create_circuit()
        self.mixer.create_circuit()
        self.initialstate.create_circuit()

        if self.anc:
            self.anc = self.problem.num_aux

        a = AncillaRegister(self.anc)
        q = QuantumRegister(self.problem.N_qubits)
        c = ClassicalRegister(self.problem.N_qubits)
        self.parameterized_circuit = QuantumCircuit(q, c, a)

        self.gamma_params = [None] * depth
        self.beta_params = [None] * depth

        self.parameterized_circuit.compose(self.initialstate.circuit, inplace=True)

        if self.usebarrier:
            self.circuit.barrier()

        for d in range(depth):
            self.gamma_params[d] = Parameter("gamma_" + str(d))
            tmp_circuit = self.problem.circuit.assign_parameters(
                {self.problem.circuit.parameters[0]: self.gamma_params[d]},
                inplace=False,
            )
            self.parameterized_circuit.compose(tmp_circuit, inplace=True)

            if self.usebarrier:
                self.circuit.barrier()

            self.beta_params[d] = Parameter("beta_" + str(d))
            tmp_circuit = self.mixer.circuit.assign_parameters(
                {self.mixer.circuit.parameters[0]: self.beta_params[d]},
                inplace=False,
            )
            self.parameterized_circuit.compose(tmp_circuit, inplace=True)

            if self.flip and (d != (depth - 1)):
                self.parameterized_circuit.barrier()
                self.flipper.create_circuit(self.bitflips[d])
                self.parameterized_circuit.compose(self.flipper.circuit, inplace=True)
                self.parameterized_circuit.barrier()

            if self.usebarrier:
                self.circuit.barrier()

        self.parameterized_circuit.barrier()
        self.parameterized_circuit.measure(q, c)
        self.parameterized_circuit = transpile(self.parameterized_circuit, self.backend)
        self.parametrized_circuit_depth = depth

    def sample_cost_landscape(
        self,
        angles={"gamma": [0, 2 * np.pi, 20], "beta": [0, 2 * np.pi, 20]},
    ):
        self.landscape_p1_angles = angles
        logger = LOG.bind(func=self.sample_cost_landscape.__name__)
        logger.info("Calculating energy landscape for depth p=1...")

        depth = 1

        tmp = angles["gamma"]
        self.gamma_grid = np.linspace(tmp[0], tmp[1], tmp[2])
        tmp = angles["beta"]
        self.beta_grid = np.linspace(tmp[0], tmp[1], tmp[2])

        if self.backend.configuration().local:
            print(depth, self.current_depth)
            self.createParameterizedCircuit(depth)
            gamma = [None] * angles["beta"][2] * angles["gamma"][2]
            beta = [None] * angles["beta"][2] * angles["gamma"][2]

            counter = 0
            for b in range(angles["beta"][2]):
                for g in range(angles["gamma"][2]):
                    gamma[counter] = self.gamma_grid[g]
                    beta[counter] = self.beta_grid[b]
                    counter += 1

            parameters = {self.gamma_params[0]: gamma, self.beta_params[0]: beta}

            logger.info("Executing sample_cost_landscape")
            logger.info(f"parameters: {len(parameters)}")
            job = self.backend.run(
                self.parameterized_circuit,
                shots=self.shots,
                parameter_binds=[parameters],
                optimization_level=0,
                memory=self.memory,
            )
            logger.info("Done execute")
            self.measurementStatistics(job)
            logger.info("Done measurement")

        else:
            raise NotImplementedError

        logger.info("Calculating Energy landscape done")

    def measurementStatistics(self, job):
        """
        implements a function for expectation value and variance

        :param job: job instance derived from BaseJob
        :return: expectation and variance, if the job is a list
        """
        jres = job.result()
        counts_list = jres.get_counts()

        if self.memorysize > 0:
            for i, _ in enumerate(jres.results):
                memory_list = jres.get_memory(experiment=i)
                if self.memorysize > 0:
                    for measurement in memory_list:
                        self.memory_lists.append(
                            [measurement, self.problem.cost(measurement[::-1])]
                        )
                        self.memorysize -= 1
                        if self.memorysize < 1:
                            break

        self.last_hist = counts_list

        if isinstance(counts_list, list):
            expectations = []
            variances = []
            maxcosts = []
            mincosts = []
            for i, counts in enumerate(counts_list):
                self.stat.reset()
                for string in counts:
                    # qiskit binary strings use little endian encoding, but our cost function expects big endian encoding. Therefore, we reverse the order
                    cost = self.problem.cost(string[::-1])
                    self.stat.add_sample(cost, counts[string], string[::-1])
                expectations.append(self.stat.get_CVaR())
                variances.append(self.stat.get_Variance())
                maxcosts.append(self.stat.get_max())
                mincosts.append(self.stat.get_min())
            angles = self.landscape_p1_angles
            self.Exp_sampled_p1 = -np.array(expectations).reshape(
                angles["beta"][2], angles["gamma"][2]
            )
            self.Var_sampled_p1 = np.array(variances).reshape(
                angles["beta"][2], angles["gamma"][2]
            )
            self.MaxCost_sampled_p1 = -np.array(maxcosts).reshape(
                angles["beta"][2], angles["gamma"][2]
            )
            self.MinCost_sampled_p1 = -np.array(mincosts).reshape(
                angles["beta"][2], angles["gamma"][2]
            )
        else:
            for string in counts_list:
                # qiskit binary strings use little endian encoding, but our cost function expects big endian encoding. Therefore, we reverse the order
                cost = self.problem.cost(string[::-1])
                self.stat.add_sample(cost, counts_list[string], string[::-1])

    def optimize(
        self,
        depth,
        angles={"gamma": [0, 2 * np.pi, 20], "beta": [0, 2 * np.pi, 20]},
    ):
        ## run local optimization by iteratively increasing the depth until depth p is reached
        while self.current_depth < depth:
            if self.current_depth == 0:
                if self.Exp_sampled_p1 is None:
                    self.sample_cost_landscape(angles=angles)
                ind_Emin = np.unravel_index(
                    np.argmin(self.Exp_sampled_p1, axis=None), self.Exp_sampled_p1.shape
                )
                angles0 = np.array(
                    (self.gamma_grid[ind_Emin[1]], self.beta_grid[ind_Emin[0]])
                )
            else:
                gamma = self.get_gamma(self.current_depth)
                beta = self.get_beta(self.current_depth)

                if self.interpolate:
                    gamma_interp = self.interp(gamma)
                    beta_interp = self.interp(beta)
                else:
                    gamma_interp = np.append(gamma, 0)
                    beta_interp = np.append(beta, 0)
                angles0 = np.zeros(2 * (self.current_depth + 1))
                angles0[::2] = gamma_interp
                angles0[1::2] = beta_interp

            self.optimization_results[self.current_depth + 1] = OptResult(
                self.current_depth + 1
            )
            # Create parameterized circuit at new depth
            self.createParameterizedCircuit(int(len(angles0) / 2))

            res = self.local_opt(angles0)

            self.samplecount_hists[self.current_depth + 1] = self.last_hist

            self.optimization_results[self.current_depth + 1].compute_best_index()

            LOG.info(
                f"cost(depth { self.current_depth + 1} = {res.fun}",
                func=self.optimize.__name__,
            )

            if self.flip:
                hist = self.samplecount_hists[self.current_depth + 1]
                string = max(hist, key=hist.get)
                boosted = self.flipper.boost_samples(
                    problem=self.problem,
                    string=string,
                )
                string_xor = self.flipper.xor(string, boosted)
                self.bitflips.append(string_xor)

            self.current_depth += 1

        if self.post:
            samples = self.samplecount_hists[self.current_depth]
            post_processing(self, samples=samples, K=self.post)
            self.Exp_post_processed = -self.stat.get_CVaR()
            self.Var_post_processed = self.stat.get_Variance()

    def local_opt(self, angles0):
        """

        :param angles0: initial guess
        """

        try:
            opt = self.optimizer[0](**self.optimizer[1])
        except TypeError as e:  ### QNSPSA needs fidelity
            self.isQNSPSA = True
            self.optimizer[1]["fidelity"] = self.optimizer[0].get_fidelity(
                self.parameterized_circuit, sampler=Sampler()
            )
            opt = self.optimizer[0](**self.optimizer[1])
        res = opt.minimize(self.loss, x0=angles0)
        if self.isQNSPSA:
            self.optimizer[1].pop("fidelity")
        return res

    def loss(self, angles):
        """
        loss function
        :return: an instance of the qiskit class QuantumCircuit
        """

        circuit = None
        n_target = self.shots
        self.stat.reset()
        shots_taken = 0
        shots = self.shots

        for i in range(3):  # this loop is used only used if precision is set
            if self.backend.configuration().local:
                params = self.getParametersToBind(
                    angles, self.parametrized_circuit_depth, asList=True
                )
                job = self.backend.run(
                    self.parameterized_circuit,
                    noise_model=self.noisemodel,
                    shots=shots,
                    parameter_binds=[params],
                    optimization_level=0,
                    memory=self.memory,
                )
            else:
                raise NotImplementedError
            shots_taken += shots
            self.measurementStatistics(job)
            if self.precision is None:
                break
            else:
                v = self.stat.get_Variance()
                shots = int((np.sqrt(v) / self.precision) ** 2) - shots_taken
                if shots <= 0:
                    break

        self.optimization_results[self.current_depth + 1].add_iteration(
            angles.copy(), self.stat, shots_taken
        )

        return -self.stat.get_CVaR()

    def getParametersToBind(self, angles, depth, asList=False):
        """
        Utility function to structure the parameterized parameter values
        so that they can be applied/bound to the parameterized circuit.

        :param angles: gamma and beta values
        :param depth: circuit depth
        :asList: Boolean that specify if the values in the dict should be a list or not
        :return: A dict containing parameters as keys and parameter values as values
        """
        assert len(angles) == 2 * depth

        params = {}
        for d in range(depth):
            if asList:
                params[self.gamma_params[d]] = [angles[2 * d + 0]]
                params[self.beta_params[d]] = [angles[2 * d + 1]]
            else:
                params[self.gamma_params[d]] = angles[2 * d + 0]
                params[self.beta_params[d]] = angles[2 * d + 1]
        return params

    def interp(self, angles):
        """
        INTERP heuristic/linear interpolation for initial parameters
        when going from depth p to p+1 (https://doi.org/10.1103/PhysRevX.10.021067)

        :param angles: angles for depth p
        :return: linear interpolation of angles for depth p+1
        """
        depth = len(angles)
        tmp = np.zeros(len(angles) + 2)
        tmp[1:-1] = angles.copy()
        w = np.arange(0, depth + 1)
        return w / depth * tmp[:-1] + (depth - w) / depth * tmp[1:]

    def hist(self, angles, shots):
        depth = int(len(angles) / 2)
        self.createParameterizedCircuit(depth)

        params = self.getParametersToBind(angles, depth, asList=True)
        if self.backend.configuration().local:
            job = self.backend.run(
                self.parameterized_circuit,
                shots=shots,
                parameter_binds=[params],
                optimization_level=0,
                memory=self.memory,
            )
        else:
            raise NotImplementedError

        # Qiskit uses big endian encoding, cost function uses litle endian encoding.
        # Therefore the string is reversed before passing it to the cost function.
        hist = job.result().get_counts()
        corrected_hist = {}
        for string in hist:
            corrected_hist[string[::-1]] = hist[string]
        return corrected_hist

    def random_init(self, gamma_bounds, beta_bounds, depth):
        """
        Enforces the bounds of gamma and beta based on the graph type.
        :param gamma_bounds: Parameter bound tuple (min,max) for gamma
        :param beta_bounds: Parameter bound tuple (min,max) for beta
        :return: np.array on the form (gamma_1, beta_1, gamma_2, ...., gamma_d, beta_d)
        """
        gamma_list = np.random.uniform(gamma_bounds[0], gamma_bounds[1], size=depth)
        beta_list = np.random.uniform(beta_bounds[0], beta_bounds[1], size=depth)
        initial = np.empty((gamma_list.size + beta_list.size,), dtype=gamma_list.dtype)
        initial[0::2] = gamma_list
        initial[1::2] = beta_list
        return initial

    def get_optimal_solutions(self):
        """
        Iterates through all optimization result objects looking for the bit-string(s)
        that gave optimal cost value.
        :return: list with the obtained best bit-strings
        """
        best_sols = []
        best_costs = []
        for i in self.optimization_results:
            best_sols_i, best_cost_i = self.optimization_results[i].get_best_solution()
            best_sols.append(best_sols_i)
            best_costs.append(best_cost_i)

        best_iterations = np.where(best_costs == np.min(best_costs))[0]
        opt_sols = []
        for i in best_iterations:
            opt_sols.append(best_sols[i])
        opt_sols = [item for sublist in opt_sols for item in sublist]
        return np.unique(opt_sols)



import numpy as np


class Statistic:
    """
    See: https://fanf2.user.srcf.net/hermes/doc/antiforgery/stats.pdf
    """

    def __init__(self, cvar=1):
        self.cvar = cvar
        self.reset()

    def reset(self):
        self.W = 0
        self.maxval = float("-inf")
        self.minval = float("inf")
        self.minSols = []
        self.maxSols = []
        self.E = 0
        self.S = 0
        self.all_values = np.array([])

    def add_sample(self, value, weight, string):
        self.W += weight
        tmp_E = self.E
        if value >= self.maxval:
            if value == self.maxval:
                self.maxSols.append(string)
            else:
                self.maxval = value
                self.maxSols = [string]
        if value <= self.minval:
            if value == self.minval:
                self.minSols.append(string)
            else:
                self.minval = value
                self.minSols = [string]

        self.maxval = max(value, self.maxval)
        self.minval = min(value, self.minval)
        self.E += weight / self.W * (value - self.E)
        self.S += weight * (value - tmp_E) * (value - self.E)
        if self.cvar < 1:
            idx = np.searchsorted(self.all_values, value)
            self.all_values = np.insert(
                self.all_values, idx, np.ones(int(weight)) * value
            )

    def get_E(self):
        return self.E

    def get_Variance(self):
        return self.S / (self.W - 1)

    def get_max(self):
        return self.maxval

    def get_min(self):
        return self.minval

    def get_max_sols(self):
        return self.maxSols

    def get_min_sols(self):
        return self.minSols

    def get_CVaR(self):
        if self.cvar < 1:
            cvarK = int(np.round(self.cvar * len(self.all_values)))
            cvar = np.sum(self.all_values[-cvarK:]) / cvarK
            return cvar
        else:
            return self.get_E()



import matplotlib.pyplot as pl
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.ticker import MaxNLocator

import numpy as np

from qaoa import QAOA
from qaoa.mixers.constrained_mixer import Constrained

from qaoa.util import Statistic

def __plot_landscape(A, extent, fig):
    if not fig:
        fig = pl.figure(figsize=(6, 6), dpi=80, facecolor="w", edgecolor="k")
    _ = pl.xlabel(r"$\gamma$")
    _ = pl.ylabel(r"$\beta$")
    ax = fig.gca()
    _ = pl.title("Expectation value")
    im = ax.imshow(A, interpolation="bicubic", origin="lower", extent=extent)
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    _ = pl.colorbar(im, cax=cax)


def plot_E(qaoa_instance, fig=None):
    angles = qaoa_instance.landscape_p1_angles
    extent = [
        angles["gamma"][0],
        angles["gamma"][1],
        angles["beta"][0],
        angles["beta"][1],
    ]
    return __plot_landscape(qaoa_instance.exp_landscape(), extent, fig=fig)


def plot_Var(qaoa_instance, fig=None):
    angles = qaoa_instance.landscape_p1_angles
    extent = [
        angles["gamma"][0],
        angles["gamma"][1],
        angles["beta"][0],
        angles["beta"][1],
    ]
    return __plot_landscape(qaoa_instance.var_landscape(), extent, fig=fig)


def plot_ApproximationRatio(
    qaoa_instance, maxdepth, mincost, maxcost, label, style="", fig=None, shots=None
):
    if not shots:
        exp = np.array(qaoa_instance.get_Exp())
    else:
        exp = []
        for p in range(1, qaoa_instance.current_depth + 1):
            ar, sp = __apprrat_successprob(qaoa_instance, p, shots=shots)
            exp.append(ar)
        exp = np.array(exp)

    if not fig:
        ax = pl.figure().gca()
    else:
        ax = fig.gca()
    pl.hlines(1, 1, maxdepth, linestyles="solid", colors="black")
    pl.plot(
        np.arange(1, maxdepth + 1),
        (maxcost - exp) / (maxcost - mincost),
        style,
        label=label,
    )
    pl.ylim(0, 1.01)
    pl.xlim(1 - 0.25, maxdepth + 0.25)
    _ = pl.ylabel("appr. ratio")
    _ = pl.xlabel("depth")
    _ = pl.legend(loc="lower right", framealpha=1)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))


def plot_successprob(
    qaoa_instance, maxdepth, label, style="", fig=None, shots=10**4
):
    successp = []
    for p in range(1, qaoa_instance.current_depth + 1):
        ar, sp = __apprrat_successprob(qaoa_instance, p, shots=shots)
        successp.append(sp)
    successp = np.array(successp)

    if not fig:
        ax = pl.figure().gca()
    else:
        ax = fig.gca()
    pl.hlines(1, 1, maxdepth, linestyles="solid", colors="black")
    pl.plot(
        np.arange(1, maxdepth + 1),
        successp,
        style,
        label=label,
    )
    pl.ylim(0, 1.01)
    pl.xlim(1 - 0.25, maxdepth + 0.25)
    _ = pl.ylabel("success prob")
    _ = pl.xlabel("depth")
    _ = pl.legend(loc="lower right", framealpha=1)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))


def __apprrat_successprob(qaoa_instance, depth, shots=10**4):
    """
    approximation ratio post processed with feasibility and success probability
    """
    hist = qaoa_instance.hist(
        qaoa_instance.optimization_results[depth].get_best_angles(), shots=shots
    )

    counts = 0

    stat = Statistic(cvar=qaoa_instance.cvar)

    for string in hist:
        if qaoa_instance.problem.isFeasible(string):
            cost = qaoa_instance.problem.cost(string)
            counts += hist[string]
            stat.add_sample(cost, hist[string], string)

    return -stat.get_CVaR(), counts / shots


def plot_angles(qaoa_instance, depth, label, style="", fig=None):
    angles=qaoa_instance.optimization_results[depth].get_best_angles()

    if not fig:
        ax = pl.figure().gca()
    else:
        ax = fig.gca()

    pl.plot(
        np.arange(1, depth + 1),
        angles[::2],
        "--"+style,
        label=r"$\gamma$ " + label,
    )
    pl.plot(
        np.arange(1, depth + 1),
        angles[1::2],
        "-"+style,
        label=r"$\beta$ " + label,
    )
    pl.xlim(1 - 0.25, depth + 0.25)
    _ = pl.ylabel("parameter")
    _ = pl.xlabel("depth")
    _ = pl.legend()
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))



import math

import numpy as np

from .base_problem import Problem
from qiskit import QuantumCircuit, QuantumRegister

from qiskit.circuit import Parameter


class ExactCover(Problem):
    def __init__(
        self,
        columns,
        weights=None,
        penalty_factor=1,
    ) -> None:
        self.columns = columns 
        self.weights = weights 
        self.penalty_factor = penalty_factor

        colSize = columns.shape[0]  ### Size per column
        numColumns = columns.shape[1]  ### number of columns/qubits

        self.N_qubits = numColumns

    def cost(self, string):
        x = np.array(list(map(int, string)))
        c_e = self.__exactCover(x)

        if self.weights is None:
            return -c_e
        else:
            return -(self.weights @ x + self.penalty_factor * c_e)

    def create_circuit(self):
        """
        Creates parameterized circuit corresponding to the cost function
        """
        q = QuantumRegister(self.N_qubits)
        self.circuit = QuantumCircuit(q)
        cost_param = Parameter("x_gamma")

        colSize, numColumns = np.shape(self.columns) 

        ### cost Hamiltonian
        for col in range(numColumns):
            hr = self.penalty_factor * 0.5 * self.columns[:, col] @ (np.sum(self.columns, axis=1) - 2)
            if not self.weights is None:
                hr += 0.5 * self.weights[col]

            if not math.isclose(hr, 0, abs_tol=1e-7):
                self.circuit.rz(cost_param * hr, q[col])

            for col_ in range(col + 1, numColumns):
                Jrr_ = self.penalty_factor * 0.5 * self.columns[:, col] @ self.columns[:, col_]

                if not math.isclose(Jrr_, 0, abs_tol=1e-7):
                    self.circuit.cx(q[col], q[col_])
                    self.circuit.rz(cost_param * Jrr_, q[col_])
                    self.circuit.cx(q[col], q[col_])

    def isFeasible(self, string):
        x = np.array(list(map(int, string)))
        c_e = self.__exactCover(x)
        return math.isclose(c_e, 0, abs_tol=1e-7)

    def __exactCover(self, x):
        return np.sum((1 - (self.columns @ x)) ** 2)



from setuptools import setup, find_packages

setup(
    name="qaoa",
    version="1.1.0",
    license="GNU General Public License v3.0",
    author="Franz Georg Fuchs",
    author_email="franzgeorgfuchs@gmail.com",
    description="Quantum Alternating Operator Ansatz/Quantum Approximate Optimization Algorithm (QAOA)",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["examples", "images"]),
    keywords="quantum computing, qaoa, qiskit",
    install_requires=[
        "numpy",
        "scipy",
        "structlog",
        "matplotlib",
        "networkx",
        "jupyter",
        "qiskit==1.1.1",
        "qiskit-aer",
        "qiskit-algorithms",
        "qiskit-finance",
        "pylatexenc",
    ],
)



import numpy as np

from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Parameter
from qiskit.circuit.library import RYGate

from .base_initialstate import InitialState


class Dicke(InitialState):
    def __init__(self, k) -> None:
        """
        Args:
            k (int): The Hamming weight of the Dicke states.
        """
        super().__init__()
        self.k = k

    def create_circuit(self):
        """
        Circuit to prepare Dicke states, following the algorithm from https://arxiv.org/pdf/1904.07358.pdf.
        """

        q = QuantumRegister(self.N_qubits)
        self.circuit = QuantumCircuit(q)

        self.circuit.x(q[-self.k :])

        for l in range(self.k + 1, self.N_qubits + 1)[::-1]:
            self.circuit.append(
                Dicke.getBlock1(self.N_qubits, self.k, l), range(self.N_qubits)
            )

        for l in range(2, self.k + 1)[::-1]:
            self.circuit.append(
                Dicke.getBlock2(self.N_qubits, self.k, l), range(self.N_qubits)
            )

    @staticmethod
    def getRYi(n):
        """
        Returns gate (i) from section 2.2.

        Args:
            n (int): The integer parameter for gate (i).

        Returns:
            QuantumCircuit: Quantum circuit representing gate (i).
        """

        qc = QuantumCircuit(2)

        qc.cx(0, 1)
        theta = 2 * np.arccos(np.sqrt(1 / n))
        ry = RYGate(theta).control(ctrl_state="1")
        qc.append(ry, [1, 0])
        qc.cx(0, 1)

        return qc

    @staticmethod
    def getRYii(l, n):
        """
        Returns gate (ii)_l from section 2.2.

        Args:
            l (int): The integer parameter for gate (ii)_l.
            n (int): The integer parameter for gate (ii)_l.

        Returns:
            QuantumCircuit: Quantum circuit representing gate (ii)_l.
        """

        qc = QuantumCircuit(3)

        qc.cx(0, 2)
        theta = 2 * np.arccos(np.sqrt(l / n))
        ry = RYGate(theta).control(num_ctrl_qubits=2, ctrl_state="11")
        qc.append(ry, [2, 1, 0])
        qc.cx(0, 2)

        return qc

    @staticmethod
    def getSCS(n, k):
        """
        Returns SCS_{n,k} gate from definition 3.

        Args:
            n (int): The integer parameter for SCS_{n,k}.
            k (int): The integer parameter for SCS_{n,k}.

        Returns:
            QuantumCircuit: Quantum circuit representing SCS_{n,k}.
        """

        qc = QuantumCircuit(k + 1)

        qc.append(Dicke.getRYi(n), [k - 1, k])
        for l in range(2, k + 1):
            qc.append(Dicke.getRYii(l, n), [k - l, k - l + 1, k])

        return qc

    @staticmethod
    def getBlock1(n, k, l):
        """
        Returns the first block in Lemma 2.

        Args:
            n (int): The integer parameter for the quantum register size.
            k (int): The integer parameter for the Hamming weight.
            l (int): The integer parameter for the block.

        Returns:
            QuantumCircuit: Quantum circuit representing the first block in Lemma 2.
        """

        qr = QuantumRegister(n)
        qc = QuantumCircuit(qr)

        first = l - k - 1
        last = n - l

        index = list(range(n))

        if first != 0:
            index = index[first:]

        if last != 0:
            index = index[:-last]
            qc.append(Dicke.getSCS(l, k), index)
        else:
            qc.append(Dicke.getSCS(l, k), index)

        return qc

    @staticmethod
    def getBlock2(n, k, l):
        """
        Returns the second block from Lemma 2.

        Args:
            n (int): The integer parameter for the quantum register size.
            k (int): The integer parameter for the Hamming weight.
            l (int): The integer parameter for the block.

        Returns:
            QuantumCircuit: Quantum circuit representing the second block in Lemma 2.
        """

        qr = QuantumRegister(n)
        qc = QuantumCircuit(qr)

        last = n - l
        index = list(range(n))

        if last != 0:
            index = index[:-last]
            qc.append(Dicke.getSCS(l, l - 1), index)
        else:
            qc.append(Dicke.getSCS(l, l - 1), index)

        return qc
